aboutsummaryrefslogtreecommitdiffhomepage
diff options
context:
space:
mode:
authorGravatar Peter Hawkins <phawkins@google.com>2017-02-27 13:02:11 -0800
committerGravatar TensorFlower Gardener <gardener@tensorflow.org>2017-02-27 14:02:18 -0800
commitb436f4130b54f0f422774d06f9affac417b9363e (patch)
tree11bbea748671ef450bf8a6a2ea88f97b1bdeae0b
parent203a4d98d696c44214854df68b43f7bd7c89ca5f (diff)
[TF:XLA] Improvements to resource variables:
* enable compilation of VarIsInitializedOp. * fix deprecated variable initializer in variable_ops_test.py * simplify variable logic in XlaContext, move intelligence into XlaOpKernelContext. * add resource variable support in the contrib layers library. Cleanups and refactorings: * merge XlaCompiler::CompileSubComputation with XlaCompiler::CompileFunction. * pass XlaCompiler arguments consistently via XlaCompiler::Options. * split the two roles of XlaCompiler::CompilationResult::input_shapes into input_mapping and xla_input_shapes. * initialize the numpy and Python seeds to a constant for XLA test cases. Change: 148683645
-rw-r--r--tensorflow/compiler/aot/compile.cc3
-rw-r--r--tensorflow/compiler/jit/kernels/xla_device_launch_op.cc31
-rw-r--r--tensorflow/compiler/jit/kernels/xla_device_launch_op.h9
-rw-r--r--tensorflow/compiler/jit/kernels/xla_local_launch_op.cc4
-rw-r--r--tensorflow/compiler/jit/xla_compilation_cache.cc1
-rw-r--r--tensorflow/compiler/jit/xla_compilation_cache.h1
-rw-r--r--tensorflow/compiler/jit/xla_device_ops.h7
-rw-r--r--tensorflow/compiler/tests/variable_ops_test.py4
-rw-r--r--tensorflow/compiler/tests/xla_test.py7
-rw-r--r--tensorflow/compiler/tf2xla/kernels/retval_op.cc2
-rw-r--r--tensorflow/compiler/tf2xla/kernels/variable_ops.cc11
-rw-r--r--tensorflow/compiler/tf2xla/op_registrations.cc2
-rw-r--r--tensorflow/compiler/tf2xla/xla_compiler.cc222
-rw-r--r--tensorflow/compiler/tf2xla/xla_compiler.h145
-rw-r--r--tensorflow/compiler/tf2xla/xla_compiler_test.cc13
-rw-r--r--tensorflow/compiler/tf2xla/xla_context.cc30
-rw-r--r--tensorflow/compiler/tf2xla/xla_context.h29
-rw-r--r--tensorflow/compiler/tf2xla/xla_op_kernel.cc58
-rw-r--r--tensorflow/compiler/tf2xla/xla_op_kernel.h8
-rw-r--r--tensorflow/contrib/framework/python/ops/variables.py12
-rw-r--r--tensorflow/contrib/layers/python/layers/layers.py5
-rw-r--r--tensorflow/python/framework/function_test.py4
-rw-r--r--tensorflow/python/ops/variable_scope.py2
23 files changed, 324 insertions, 286 deletions
diff --git a/tensorflow/compiler/aot/compile.cc b/tensorflow/compiler/aot/compile.cc
index 6b2cf451f5..1284155c07 100644
--- a/tensorflow/compiler/aot/compile.cc
+++ b/tensorflow/compiler/aot/compile.cc
@@ -298,8 +298,7 @@ Status ConvertGraphToXla(xla::LocalClient* client, std::unique_ptr<Graph> graph,
graph->versions().producer(), flib_def, OptimizerOptions()));
XlaCompiler::CompilationResult result;
TF_RETURN_IF_ERROR(compiler.CompileGraph("tfcompile", std::move(graph),
- flib_run.get(), xla_args,
- false /* use_tuple_arg */, &result));
+ flib_run.get(), xla_args, &result));
*has_context_arg = result.requires_runtime_context;
*computation = std::move(result.computation);
diff --git a/tensorflow/compiler/jit/kernels/xla_device_launch_op.h b/tensorflow/compiler/jit/kernels/xla_device_launch_op.h
index c77d5323b5..65516163c9 100644
--- a/tensorflow/compiler/jit/kernels/xla_device_launch_op.h
+++ b/tensorflow/compiler/jit/kernels/xla_device_launch_op.h
@@ -16,6 +16,7 @@ limitations under the License.
#ifndef TENSORFLOW_COMPILER_JIT_KERNELS_XLA_DEVICE_LAUNCH_OP_H_
#define TENSORFLOW_COMPILER_JIT_KERNELS_XLA_DEVICE_LAUNCH_OP_H_
+#include "tensorflow/compiler/jit/xla_compilation_cache.h"
#include "tensorflow/core/framework/allocator.h"
#include "tensorflow/core/framework/op.h"
#include "tensorflow/core/framework/op_kernel.h"
@@ -24,6 +25,14 @@ limitations under the License.
namespace tensorflow {
+// Takes a snapshot of the values of resource variable arguments, which are
+// the last `num_variables` arguments. We snapshot tensors that back
+// resource variables since concurrent updates may modify the shape, and it is
+// important that the shapes used for compilation match the true shapes of the
+// buffers.
+std::vector<OptionalTensor> SnapshotResourceVariables(OpKernelContext* ctx,
+ int num_variables);
+
// The XlaDeviceLaunchOp is used to replace a region of the TensorFlow graph
// which will be compiled and executed using XLA. The XlaDeviceLaunchOp is
// responsible for handling interactions with the TensorFlow executor.
diff --git a/tensorflow/compiler/jit/kernels/xla_local_launch_op.cc b/tensorflow/compiler/jit/kernels/xla_local_launch_op.cc
index e056442975..5bcb6b0b60 100644
--- a/tensorflow/compiler/jit/kernels/xla_local_launch_op.cc
+++ b/tensorflow/compiler/jit/kernels/xla_local_launch_op.cc
@@ -219,8 +219,8 @@ void XlaLocalLaunchOp::Compute(OpKernelContext* ctx) {
// Pass remaining parameters.
for (int i = 0; i < kernel->xla_input_shapes.size(); ++i) {
- int arg_num = kernel->xla_input_shapes[i].first;
- const xla::Shape& shape = kernel->xla_input_shapes[i].second;
+ int arg_num = kernel->input_mapping[i];
+ const xla::Shape& shape = kernel->xla_input_shapes[i];
gpu::DeviceMemoryBase dmem(
const_cast<char*>(ctx->input(arg_num).tensor_data().data()),
ctx->input(arg_num).tensor_data().size());
diff --git a/tensorflow/compiler/jit/xla_compilation_cache.cc b/tensorflow/compiler/jit/xla_compilation_cache.cc
index 32e706c50f..41abea02eb 100644
--- a/tensorflow/compiler/jit/xla_compilation_cache.cc
+++ b/tensorflow/compiler/jit/xla_compilation_cache.cc
@@ -181,6 +181,7 @@ Status BuildArguments(int num_constant_args,
XlaCompiler::Argument& arg = (*args)[input_num];
+ arg.name = variable_args[variable_id].name;
if (variable_args[variable_id].present) {
const Tensor& value = variable_args[variable_id].value;
arg.kind = XlaCompiler::Argument::kVariable;
diff --git a/tensorflow/compiler/jit/xla_compilation_cache.h b/tensorflow/compiler/jit/xla_compilation_cache.h
index 2f311b961c..ff67e48d1a 100644
--- a/tensorflow/compiler/jit/xla_compilation_cache.h
+++ b/tensorflow/compiler/jit/xla_compilation_cache.h
@@ -31,6 +31,7 @@ namespace tensorflow {
// Struct that represents a possibly-absent Tensor.
struct OptionalTensor {
+ string name; // A descriptive name
bool present = false; // Is the tensor present?
Tensor value; // If present, what is the Tensor's value?
};
diff --git a/tensorflow/compiler/jit/xla_device_ops.h b/tensorflow/compiler/jit/xla_device_ops.h
index 7a0a212f5a..b084dcaa7d 100644
--- a/tensorflow/compiler/jit/xla_device_ops.h
+++ b/tensorflow/compiler/jit/xla_device_ops.h
@@ -112,12 +112,7 @@ class XlaDeviceDummyOp : public OpKernel {
\
REGISTER_KERNEL_BUILDER( \
Name("VarHandleOp").Device(DEVICE).HostMemory("resource"), \
- ResourceHandleOp<Var>); \
- REGISTER_KERNEL_BUILDER(Name("VarIsInitializedOp") \
- .Device(DEVICE) \
- .HostMemory("resource") \
- .HostMemory("is_initialized"), \
- IsResourceInitialized<Var>);
+ ResourceHandleOp<Var>);
// TODO(b/32507444): the registrations for the control flow operators are
// temporary and exist primarily to work around a bug in the graph partitioning
diff --git a/tensorflow/compiler/tests/variable_ops_test.py b/tensorflow/compiler/tests/variable_ops_test.py
index f68e1f9fbc..dcb9e2db2f 100644
--- a/tensorflow/compiler/tests/variable_ops_test.py
+++ b/tensorflow/compiler/tests/variable_ops_test.py
@@ -56,7 +56,7 @@ class VariableOpsTest(XLATestCase):
with ops.control_dependencies([d]):
e = x.read_value()
- session.run(variables.initialize_all_variables())
+ session.run(variables.global_variables_initializer())
v1, v2, v3 = session.run([a, c, e])
self.assertAllClose(2.0, v1)
self.assertAllClose(47.0, v2)
@@ -86,7 +86,7 @@ class VariableOpsTest(XLATestCase):
optimizer = GradientDescentOptimizer(0.1)
train = optimizer.minimize(loss)
- session.run(variables.initialize_all_variables())
+ session.run(variables.global_variables_initializer())
session.run(train, {x: np.array([[7, 3, 5, 9]], dtype=np.float32)})
vw, vb = session.run([w, b])
self.assertAllClose(
diff --git a/tensorflow/compiler/tests/xla_test.py b/tensorflow/compiler/tests/xla_test.py
index b72e7c9713..dfb4904338 100644
--- a/tensorflow/compiler/tests/xla_test.py
+++ b/tensorflow/compiler/tests/xla_test.py
@@ -19,14 +19,18 @@ from __future__ import division
from __future__ import print_function
import contextlib
+import random
import re
+import numpy as np
+
from tensorflow.contrib.compiler import jit
from tensorflow.core.framework import types_pb2
from tensorflow.core.protobuf import config_pb2
from tensorflow.python.client import session
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
+from tensorflow.python.framework import random_seed
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import variables
from tensorflow.python.platform import flags
@@ -81,6 +85,9 @@ class XLATestCase(test.TestCase):
return
logging.info('Start test case: %s', name)
+ random.seed(random_seed.DEFAULT_GRAPH_SEED)
+ np.random.seed(random_seed.DEFAULT_GRAPH_SEED)
+
def tearDown(self):
logging.info('End test case: %s', self._testMethodName)
diff --git a/tensorflow/compiler/tf2xla/kernels/retval_op.cc b/tensorflow/compiler/tf2xla/kernels/retval_op.cc
index 72ab1249f9..ae9cecc10b 100644
--- a/tensorflow/compiler/tf2xla/kernels/retval_op.cc
+++ b/tensorflow/compiler/tf2xla/kernels/retval_op.cc
@@ -60,7 +60,7 @@ class RetvalOp : public XlaOpKernel {
OP_REQUIRES_OK(ctx, ctx->ConstantInput(0, &literal));
OP_REQUIRES_OK(ctx, tc.AddConstRetval(index_, dtype_, literal));
} else {
- tc.AddRetval(index_, input);
+ tc.AddRetval(index_, dtype_, input);
}
}
}
diff --git a/tensorflow/compiler/tf2xla/kernels/variable_ops.cc b/tensorflow/compiler/tf2xla/kernels/variable_ops.cc
index ee984cd119..f7326b0edd 100644
--- a/tensorflow/compiler/tf2xla/kernels/variable_ops.cc
+++ b/tensorflow/compiler/tf2xla/kernels/variable_ops.cc
@@ -25,6 +25,17 @@ limitations under the License.
namespace tensorflow {
namespace {
+class VarIsInitializedOp : public XlaOpKernel {
+ public:
+ explicit VarIsInitializedOp(OpKernelConstruction* ctx) : XlaOpKernel(ctx) {}
+ void Compile(XlaOpKernelContext* ctx) override {
+ xla::ComputationDataHandle handle;
+ bool initialized = ctx->ReadVariableInput(0, &handle).ok();
+ ctx->SetOutput(0, ctx->builder()->ConstantR0<bool>(initialized));
+ }
+};
+REGISTER_XLA_OP("VarIsInitializedOp", VarIsInitializedOp);
+
class ReadVariableOp : public XlaOpKernel {
public:
explicit ReadVariableOp(OpKernelConstruction* ctx) : XlaOpKernel(ctx) {}
diff --git a/tensorflow/compiler/tf2xla/op_registrations.cc b/tensorflow/compiler/tf2xla/op_registrations.cc
index 7f779a28e4..171e96d3b6 100644
--- a/tensorflow/compiler/tf2xla/op_registrations.cc
+++ b/tensorflow/compiler/tf2xla/op_registrations.cc
@@ -276,6 +276,7 @@ REGISTER_XLA_KERNEL(DEVICE_CPU_XLA_JIT,
Name("TruncateMod").TypeConstraint("T", kCpuNumericTypes));
REGISTER_XLA_KERNEL(DEVICE_CPU_XLA_JIT,
Name("Unpack").TypeConstraint("T", kCpuAllTypes));
+REGISTER_XLA_KERNEL(DEVICE_CPU_XLA_JIT, Name("VarIsInitializedOp"));
REGISTER_XLA_KERNEL(DEVICE_CPU_XLA_JIT,
Name("ZerosLike").TypeConstraint("T", kCpuNumericTypes));
@@ -536,6 +537,7 @@ REGISTER_XLA_KERNEL(DEVICE_GPU_XLA_JIT,
Name("TruncateMod").TypeConstraint("T", kGpuNumericTypes));
REGISTER_XLA_KERNEL(DEVICE_GPU_XLA_JIT,
Name("Unpack").TypeConstraint("T", kGpuAllTypes));
+REGISTER_XLA_KERNEL(DEVICE_GPU_XLA_JIT, Name("VarIsInitializedOp"));
REGISTER_XLA_KERNEL(DEVICE_GPU_XLA_JIT,
Name("ZerosLike").TypeConstraint("T", kGpuNumericTypes));
diff --git a/tensorflow/compiler/tf2xla/xla_compiler.cc b/tensorflow/compiler/tf2xla/xla_compiler.cc
index aea50bb5cd..efc8dfce93 100644
--- a/tensorflow/compiler/tf2xla/xla_compiler.cc
+++ b/tensorflow/compiler/tf2xla/xla_compiler.cc
@@ -27,6 +27,7 @@ limitations under the License.
#include "tensorflow/core/common_runtime/function.h"
#include "tensorflow/core/common_runtime/graph_optimizer.h"
#include "tensorflow/core/framework/attr_value_util.h"
+#include "tensorflow/core/graph/algorithm.h"
#include "tensorflow/core/graph/graph_constructor.h"
#include "tensorflow/core/graph/node_builder.h"
#include "tensorflow/core/lib/hash/hash.h"
@@ -37,36 +38,18 @@ namespace tensorflow {
namespace {
-Status CheckSignature(const DataTypeVector& tf_types,
- const xla::Shape& xla_shape) {
- if (xla::ShapeUtil::IsTuple(xla_shape)) {
- if (xla::ShapeUtil::TupleElementCount(xla_shape) != tf_types.size()) {
- return errors::Internal("XLA shape has ",
- xla::ShapeUtil::TupleElementCount(xla_shape),
- " elements while function has ", tf_types.size());
- }
- for (int i = 0; i < tf_types.size(); ++i) {
- xla::PrimitiveType type;
- TF_RETURN_IF_ERROR(DataTypeToPrimitiveType(tf_types[i], &type));
- if (type !=
- xla::ShapeUtil::GetTupleElementShape(xla_shape, i).element_type()) {
- return errors::Internal(
- "element ", i, " has XLA type ",
- xla::ShapeUtil::GetTupleElementShape(xla_shape, i).element_type(),
- " and TensorFlow type ", DataTypeString(tf_types[i]));
- }
- }
- } else {
- if (tf_types.size() != 1) {
- return errors::Internal("Expected singleton type, got ", tf_types.size(),
- " types");
- }
- xla::PrimitiveType type;
- TF_RETURN_IF_ERROR(DataTypeToPrimitiveType(tf_types[0], &type));
- if (type != xla_shape.element_type()) {
- return errors::Internal("singleton element has XLA type ",
- xla_shape.element_type(), " and TensorFlow type ",
- DataTypeString(tf_types[0]));
+// Checks that arguments `args` match types `types`.
+Status CheckSignature(const DataTypeVector& types,
+ const std::vector<XlaCompiler::Argument>& args) {
+ if (args.size() != types.size()) {
+ return errors::Internal("Compilation arguments have ", args.size(),
+ " elements while function has ", types.size());
+ }
+ for (int i = 0; i < types.size(); ++i) {
+ if (types[i] != args[i].type && types[i] != DT_RESOURCE) {
+ return errors::Internal(
+ "Argument ", i, " has declared type ", DataTypeString(args[i].type),
+ " but function parameter has type ", DataTypeString(types[i]));
}
}
return Status::OK();
@@ -74,14 +57,10 @@ Status CheckSignature(const DataTypeVector& tf_types,
} // namespace
-XlaCompiler::XlaCompiler(const XlaCompiler::Options& options)
- : client_(options.client),
- allow_cpu_custom_calls_(options.allow_cpu_custom_calls),
- local_executable_has_hybrid_result_(
- options.local_executable_has_hybrid_result),
- resolve_compile_time_constants_(options.resolve_compile_time_constants),
+XlaCompiler::XlaCompiler(XlaCompiler::Options options)
+ : options_(std::move(options)),
next_step_id_(1),
- device_(new XlaCompilationDevice(SessionOptions(), options.device_type)),
+ device_(new XlaCompilationDevice(SessionOptions(), options_.device_type)),
device_mgr_({device_}) {}
XlaCompiler::~XlaCompiler() = default;
@@ -91,6 +70,19 @@ int64 XlaCompiler::NextStepId() {
return next_step_id_++;
}
+// Prunes any nodes from a function that are not dependencies of the _Retval
+// nodes. Used to prune stateful ops from within a function body, such as
+// variable initializers, that should not be executed unless requested.
+static void PruneUnreachableNodes(Graph* graph) {
+ std::unordered_set<const Node*> nodes;
+ for (Node* node : graph->nodes()) {
+ if (node->type_string() == "_Retval") {
+ nodes.insert(node);
+ }
+ }
+ PruneForReverseReachability(graph, nodes);
+}
+
Status XlaCompiler::CompileFunction(
FunctionLibraryRuntime* flr, const NameAttrList& function,
const std::vector<XlaCompiler::Argument>& args,
@@ -105,69 +97,14 @@ Status XlaCompiler::CompileFunction(
const FunctionBody* fbody = flr->GetFunctionBody(handle);
CHECK(fbody);
- return CompileFunctionBody(flr, *fbody, function_id, args,
- /*use_tuple_arg=*/false, result);
-}
-
-Status XlaCompiler::CompileSubComputation(FunctionLibraryRuntime* flr,
- const NameAttrList& function,
- const xla::Shape& input_shape,
- const xla::Shape& output_shape,
- xla::Computation* computation) {
- const string function_id = Canonicalize(function.name(), function.attr());
- VLOG(1) << "XlaCompiler::CompileSubComputation " << function_id;
-
- FunctionLibraryRuntime::Handle handle;
- TF_RETURN_IF_ERROR(
- flr->Instantiate(function.name(), function.attr(), &handle));
-
- const FunctionBody* fbody = flr->GetFunctionBody(handle);
- CHECK(fbody);
-
- TF_RETURN_IF_ERROR(CheckSignature(fbody->arg_types, input_shape));
- TF_RETURN_IF_ERROR(CheckSignature(fbody->ret_types, output_shape));
-
- const bool use_tuple_arg = xla::ShapeUtil::IsTuple(input_shape);
-
- std::vector<XlaCompiler::Argument> args(fbody->arg_types.size());
- if (use_tuple_arg) {
- for (int i = 0; i < args.size(); ++i) {
- xla::Shape xla_shape =
- xla::ShapeUtil::GetTupleElementShape(input_shape, i);
- args[i].kind = Argument::kParameter;
- args[i].type = fbody->arg_types[i];
- args[i].shape = XLAShapeToTensorShape(xla_shape);
- }
- } else {
- args[0].kind = Argument::kParameter;
- args[0].type = fbody->arg_types[0];
- args[0].shape = XLAShapeToTensorShape(input_shape);
- }
-
- CompilationResult result;
- TF_RETURN_IF_ERROR(CompileFunctionBody(flr, *fbody, function_id, args,
- use_tuple_arg, &result));
-
- if (!xla::ShapeUtil::Compatible(result.xla_output_shape, output_shape)) {
- return errors::Internal("output shape mismatch from compilation");
- }
- *computation = std::move(result.computation);
-
- return Status::OK();
-}
-
-Status XlaCompiler::CompileFunctionBody(
- FunctionLibraryRuntime* flr, const FunctionBody& fbody,
- const string& function_id, const std::vector<XlaCompiler::Argument>& args,
- bool use_tuple_arg, XlaCompiler::CompilationResult* result) {
- VLOG(1) << "XlaCompiler::CompileFunctionBody " << function_id;
+ TF_RETURN_IF_ERROR(CheckSignature(fbody->arg_types, args));
std::unique_ptr<Graph> graph(new Graph(flr->GetFunctionLibraryDefinition()));
- CopyGraph(*fbody.graph, graph.get());
+ CopyGraph(*fbody->graph, graph.get());
if (VLOG_IS_ON(1)) {
dump_graph::DumpGraphToFile(
- strings::StrCat("xla_jit_raw_input_", function_id), *graph);
+ strings::StrCat("xla_compile_function_input_", function_id), *graph);
}
// Optimize the graph before running the compiler.
@@ -179,12 +116,13 @@ Status XlaCompiler::CompileFunctionBody(
if (VLOG_IS_ON(1)) {
dump_graph::DumpGraphToFile(
- strings::StrCat("xla_jit_final_graph_", function_id), *graph);
+ strings::StrCat("xla_compile_function_optimized_", function_id),
+ *graph);
}
VLOG(1) << "====================================================";
- TF_RETURN_IF_ERROR(CompileGraph(function_id, std::move(graph), flr, args,
- use_tuple_arg, result));
+ TF_RETURN_IF_ERROR(
+ CompileGraph(function_id, std::move(graph), flr, args, result));
VLOG(1) << "====================================================";
return Status::OK();
@@ -199,7 +137,7 @@ Status XlaCompiler::BuildExecutable(
std::vector<const xla::Shape*> argument_layouts(
result.xla_input_shapes.size());
for (int i = 0; i < result.xla_input_shapes.size(); ++i) {
- argument_layouts[i] = &result.xla_input_shapes[i].second;
+ argument_layouts[i] = &result.xla_input_shapes[i];
}
if (result.requires_runtime_context) {
// The final arg is the XlaLocalRuntimeContext*.
@@ -210,7 +148,8 @@ Status XlaCompiler::BuildExecutable(
build_options.set_device_ordinal(local_client->default_device_ordinal());
build_options.set_platform(local_client->platform());
build_options.set_result_layout(result.xla_output_shape);
- build_options.set_has_hybrid_result(local_executable_has_hybrid_result_);
+ build_options.set_has_hybrid_result(
+ options_.local_executable_has_hybrid_result);
auto compile_result = local_client->Compile(result.computation,
argument_layouts, build_options);
@@ -272,13 +211,12 @@ Status ExecuteGraph(XlaContext* xla_context, std::unique_ptr<Graph> graph,
}
// Builds XLA computations for each of the arguments to the computation.
-// `args` are the arguments to the computation. If `use_tuple_arg` is true, a
-// single tuple parameter will be used for all arguments; if false, each
-// argument gets its own parameter.
+// `args` are the arguments to the computation.
Status BuildArguments(const std::vector<XlaCompiler::Argument>& args,
bool use_tuple_arg, xla::ComputationBuilder* builder,
std::vector<XlaContext::Argument>* context_args,
- std::vector<std::pair<int, xla::Shape>>* input_shapes) {
+ std::vector<int>* input_mapping,
+ std::vector<xla::Shape>* input_shapes) {
context_args->resize(args.size());
// Argument numbers of arguments and variables that are to be passed to the
@@ -322,31 +260,30 @@ Status BuildArguments(const std::vector<XlaCompiler::Argument>& args,
return Status::OK();
}
- std::vector<xla::Shape> parameter_shapes(parameters.size());
input_shapes->resize(parameters.size());
- for (int i = 0; i < parameters.size(); ++i) {
+ input_mapping->resize(parameters.size());
+ for (int i = 0; i < input_shapes->size(); ++i) {
const XlaCompiler::Argument& arg = args[parameters[i]];
// Computes the shapes of non-constant arguments.
xla::PrimitiveType type;
TF_RETURN_IF_ERROR(DataTypeToPrimitiveType(arg.type, &type));
xla::ShapeUtil::PopulateShape(type, arg.shape.dim_sizes(),
- &parameter_shapes[i]);
- (*input_shapes)[i].first = parameters[i];
- (*input_shapes)[i].second = parameter_shapes[i];
+ &(*input_shapes)[i]);
+ (*input_mapping)[i] = parameters[i];
}
if (use_tuple_arg) {
- xla::Shape tuple_shape = xla::ShapeUtil::MakeTupleShape(parameter_shapes);
+ xla::Shape tuple_shape = xla::ShapeUtil::MakeTupleShape(*input_shapes);
xla::ComputationDataHandle tuple =
builder->Parameter(0, tuple_shape, "arg_tuple");
- for (int i = 0; i < parameters.size(); ++i) {
+ for (int i = 0; i < input_shapes->size(); ++i) {
(*context_args)[parameters[i]].value.handle =
builder->GetTupleElement(tuple, i);
}
} else {
- for (int i = 0; i < parameters.size(); ++i) {
+ for (int i = 0; i < input_shapes->size(); ++i) {
(*context_args)[parameters[i]].value.handle =
- builder->Parameter(i, parameter_shapes[i], strings::StrCat("arg", i));
+ builder->Parameter(i, (*input_shapes)[i], strings::StrCat("arg", i));
}
}
return Status::OK();
@@ -359,19 +296,22 @@ Status BuildArguments(const std::vector<XlaCompiler::Argument>& args,
// variable states, generated by the symbolic evaluation.
// If `has_side_effects` is true, the computation has side effects and should be
// built even if it has no outputs.
+// If `return_updated_values_for_all_variables` is true, all variables will be
+// included in `variable_updates`, regardless of whether their value changed.
// Sets `*num_nonconst_outputs` to the number of outputs of the `computation`.
-// Sets `*variable_writes` to a description of variables whose values are
+// Sets `*variable_updates` to a description of variables whose values are
// written by the computation; the variable writes are the last
-// `variable_writes.size()` return values from the computation. Each entry in
-// `variable_writes` is a (input_index, type) pair, where `input_index` is the
+// `variable_updates.size()` return values from the computation. Each entry in
+// `variable_updates` is a (input_index, type) pair, where `input_index` is the
// index of a resource variable argument to the computation, and `type` is the
// type of the final output.
Status BuildComputation(
const std::vector<XlaContext::HandleOrConstant>& retvals,
const std::unordered_map<int, XlaContext::Variable>& variable_map,
- bool has_side_effects, xla::ComputationBuilder* builder,
- xla::Computation* computation, int* num_nonconst_outputs,
- std::vector<std::pair<int, DataType>>* variable_writes) {
+ bool has_side_effects, bool return_updated_values_for_all_variables,
+ xla::ComputationBuilder* builder, xla::Computation* computation,
+ int* num_nonconst_outputs,
+ std::vector<XlaCompiler::VariableUpdate>* variable_updates) {
std::vector<xla::ComputationDataHandle> elems;
elems.reserve(retvals.size());
for (const XlaContext::HandleOrConstant& retval : retvals) {
@@ -394,8 +334,14 @@ Status BuildComputation(
});
for (const auto& entry : variables) {
- if (entry.second->value.handle() != entry.second->initial_value.handle()) {
- variable_writes->emplace_back(entry.first, entry.second->type);
+ bool modified =
+ entry.second->value.handle() != entry.second->initial_value.handle();
+ if (return_updated_values_for_all_variables || modified) {
+ variable_updates->emplace_back();
+ XlaCompiler::VariableUpdate& update = variable_updates->back();
+ update.input_index = entry.first;
+ update.type = entry.second->type;
+ update.modified = modified;
elems.push_back(entry.second->value);
}
}
@@ -428,34 +374,41 @@ Status XlaCompiler::CompileGraph(string const& name,
std::unique_ptr<Graph> graph,
FunctionLibraryRuntime* flib,
const std::vector<XlaCompiler::Argument>& args,
- bool use_tuple_arg,
CompilationResult* result) {
VLOG(1) << "Executing graph symbolically to populate ComputationBuilder.";
xla::ComputationBuilder builder(client(), name);
- XlaContext* context = new XlaContext(this, &builder, allow_cpu_custom_calls_,
- resolve_compile_time_constants_);
+ XlaContext* context =
+ new XlaContext(this, &builder, options_.allow_cpu_custom_calls,
+ options_.resolve_compile_time_constants);
core::ScopedUnref context_unref(context);
+ result->tuple_arg = options_.use_tuple_arg;
+
std::vector<XlaContext::Argument> context_args;
- TF_RETURN_IF_ERROR(BuildArguments(args, use_tuple_arg, &builder,
- &context_args, &result->xla_input_shapes));
+ TF_RETURN_IF_ERROR(BuildArguments(args, options_.use_tuple_arg, &builder,
+ &context_args, &result->input_mapping,
+ &result->xla_input_shapes));
context->set_args(std::move(context_args));
+ if (options_.prune_unreachable_nodes) {
+ PruneUnreachableNodes(graph.get());
+ }
+
TF_RETURN_IF_ERROR(
ExecuteGraph(context, std::move(graph), device_, flib, NextStepId()));
int num_nonconst_outputs;
- std::vector<std::pair<int, DataType>> variable_writes;
TF_RETURN_IF_ERROR(BuildComputation(
context->retvals(), context->variables(), context->has_side_effects(),
- &builder, &result->computation, &num_nonconst_outputs, &variable_writes));
+ options_.return_updated_values_for_all_variables, &builder,
+ &result->computation, &num_nonconst_outputs, &result->variable_updates));
result->requires_runtime_context = context->has_context_parameter();
// Tuple arguments and runtime context parameters are incompatible.
- CHECK(!(use_tuple_arg && result->requires_runtime_context));
+ CHECK(!(options_.use_tuple_arg && result->requires_runtime_context));
VLOG(2) << "Outputs: total: " << context->retvals().size()
<< " nonconstant: " << num_nonconst_outputs;
@@ -521,17 +474,14 @@ Status XlaCompiler::CompileGraph(string const& name,
}
}
- result->variable_writes.resize(variable_writes.size());
- for (int i = 0; i < variable_writes.size(); ++i) {
- result->variable_writes[i].input_index = variable_writes[i].first;
- result->variable_writes[i].type = variable_writes[i].second;
+ for (int i = 0; i < result->variable_updates.size(); ++i) {
if (num_computation_outputs > 1) {
- result->variable_writes[i].shape =
+ result->variable_updates[i].shape =
XLAShapeToTensorShape(xla::ShapeUtil::GetTupleElementShape(
result->xla_output_shape, computation_output));
} else {
CHECK_EQ(0, computation_output);
- result->variable_writes[i].shape =
+ result->variable_updates[i].shape =
XLAShapeToTensorShape(result->xla_output_shape);
}
++computation_output;
@@ -544,7 +494,7 @@ Status XlaCompiler::GetChannelHandle(const string& key,
mutex_lock lock(mu_);
auto result = channels_.emplace(key, xla::ChannelHandle());
if (result.second) {
- TF_ASSIGN_OR_RETURN(result.first->second, client_->CreateChannelHandle());
+ TF_ASSIGN_OR_RETURN(result.first->second, client()->CreateChannelHandle());
}
*channel = result.first->second;
VLOG(1) << "Channel: " << key << " " << channel->DebugString();
diff --git a/tensorflow/compiler/tf2xla/xla_compiler_test.cc b/tensorflow/compiler/tf2xla/xla_compiler_test.cc
index b1b4c26b15..aa809f85a1 100644
--- a/tensorflow/compiler/tf2xla/xla_compiler_test.cc
+++ b/tensorflow/compiler/tf2xla/xla_compiler_test.cc
@@ -71,8 +71,7 @@ TEST_F(XlaCompilerTest, EmptyReturnValues) {
std::unique_ptr<Graph> graph(new Graph(OpRegistry::Global()));
XlaCompiler::CompilationResult result;
TF_ASSERT_OK(compiler.CompileGraph("add", std::move(graph), flr.get(),
- /*args=*/{}, /*use_tuple_arg=*/false,
- &result));
+ /*args=*/{}, &result));
// No computation should be generated.
EXPECT_EQ(0, result.computation.handle().handle());
@@ -103,8 +102,8 @@ TEST_F(XlaCompilerTest, Simple) {
auto flr = BuildFunctionLibraryRuntime(compiler);
XlaCompiler::CompilationResult result;
- TF_ASSERT_OK(compiler.CompileGraph("add", std::move(graph), flr.get(), args,
- /*use_tuple_arg=*/false, &result));
+ TF_ASSERT_OK(
+ compiler.CompileGraph("add", std::move(graph), flr.get(), args, &result));
// Tests that the generated computation works.
std::unique_ptr<xla::Literal> param0_literal =
@@ -160,8 +159,7 @@ TEST_F(XlaCompilerTest, ConstantOutputs) {
XlaCompiler::CompilationResult result;
TF_ASSERT_OK(compiler.CompileGraph("constants", std::move(graph_copy),
- flr.get(), args, /*use_tuple_arg=*/false,
- &result));
+ flr.get(), args, &result));
ASSERT_EQ(2, result.outputs.size());
EXPECT_TRUE(result.outputs[0].is_constant);
@@ -198,8 +196,7 @@ TEST_F(XlaCompilerTest, ConstantOutputs) {
XlaCompiler::CompilationResult result;
TF_ASSERT_OK(compiler.CompileGraph("constants", std::move(graph_copy),
- flr.get(), args, /*use_tuple_arg=*/false,
- &result));
+ flr.get(), args, &result));
ASSERT_EQ(2, result.outputs.size());
EXPECT_FALSE(result.outputs[0].is_constant);
diff --git a/tensorflow/compiler/tf2xla/xla_context.cc b/tensorflow/compiler/tf2xla/xla_context.cc
index 9af0f544e9..57d946509b 100644
--- a/tensorflow/compiler/tf2xla/xla_context.cc
+++ b/tensorflow/compiler/tf2xla/xla_context.cc
@@ -86,7 +86,7 @@ string XlaContext::DebugString() { return "TLA JIT context"; }
// This is called by the Retval Op to associate a computed value
// with a specific return value of the subgraph.
-void XlaContext::AddRetval(int retval_index,
+void XlaContext::AddRetval(int retval_index, DataType type,
const xla::ComputationDataHandle& handle) {
VLOG(1) << "Added retval index " << retval_index << " to XLA computation";
// Add the return value to the list being built up.
@@ -94,6 +94,7 @@ void XlaContext::AddRetval(int retval_index,
retvals_.resize(retval_index + 1);
}
retvals_[retval_index].is_constant = false;
+ retvals_[retval_index].type = type;
retvals_[retval_index].handle = handle;
}
@@ -104,6 +105,7 @@ Status XlaContext::AddConstRetval(int retval_index, DataType dtype,
if (retvals_.size() <= retval_index) {
retvals_.resize(retval_index + 1);
}
+ retvals_[retval_index].type = dtype;
if (resolve_compile_time_constants_) {
retvals_[retval_index].is_constant = true;
TF_RETURN_IF_ERROR(LiteralToHostTensor(
@@ -135,34 +137,12 @@ Status XlaContext::CreateVariable(int variable_id, string name, DataType type,
return Status::OK();
}
-Status XlaContext::AssignVariable(int variable_id, DataType type,
- const xla::ComputationDataHandle& handle) {
+Status XlaContext::GetVariable(int variable_id, Variable** variable) {
auto it = variables_.find(variable_id);
if (it == variables_.end()) {
return errors::InvalidArgument("Unknown variable ID ", variable_id);
}
- Variable& var = it->second;
- if (!((var.type == DT_INVALID && type != DT_INVALID) || (var.type == type))) {
- return errors::InvalidArgument(
- "Types of variables cannot change after initialization: old type was ",
- DataTypeString(var.type), ", new type is ", DataTypeString(type));
- }
- var.type = type;
- var.value = handle;
- return Status::OK();
-}
-
-Status XlaContext::ReadVariable(int variable_id,
- xla::ComputationDataHandle* handle) {
- auto it = variables_.find(variable_id);
- if (it == variables_.end()) {
- return errors::InvalidArgument("Unknown variable ID ", variable_id);
- }
- *handle = it->second.value;
- if (handle->handle() == 0) {
- return errors::InvalidArgument("Read of uninitialized variable ",
- it->second.name);
- }
+ *variable = &it->second;
return Status::OK();
}
diff --git a/tensorflow/compiler/tf2xla/xla_context.h b/tensorflow/compiler/tf2xla/xla_context.h
index 5d56eedf32..657ead5391 100644
--- a/tensorflow/compiler/tf2xla/xla_context.h
+++ b/tensorflow/compiler/tf2xla/xla_context.h
@@ -93,7 +93,8 @@ class XlaContext : public ResourceBase {
// This is called by the Retval Op to associate a computed value
// with a specific return value of the subgraph.
- void AddRetval(int retval_index, const xla::ComputationDataHandle& handle);
+ void AddRetval(int retval_index, DataType type,
+ const xla::ComputationDataHandle& handle);
// As for Retval, but for return values that are compile-time constants.
Status AddConstRetval(int retval_index, DataType dtype,
@@ -104,22 +105,6 @@ class XlaContext : public ResourceBase {
bool has_side_effects() const { return has_side_effects_; }
- // Creates a variable with variable `variable_id` and initial type `type` and
- // value `handle`. `name` is a descriptive name for use in error messages.
- // Fails if the variable already exists.
- Status CreateVariable(int variable_id, string name, DataType type,
- const xla::ComputationDataHandle& handle);
-
- // Assigns value `handle` with type `type` to variable `variable_id`. Fails if
- // the variable has not already been created using CreateVariable.
- Status AssignVariable(int variable_id, DataType type,
- const xla::ComputationDataHandle& handle);
-
- // Reads the current value of `variable_id`, setting `handle` to its current
- // value. Returns a failure status if the variable has not been created or
- // its value has not been initialized.
- Status ReadVariable(int variable_id, xla::ComputationDataHandle* handle);
-
struct Variable {
// A descriptive name for the variable, used in error messages.
string name;
@@ -136,6 +121,16 @@ class XlaContext : public ResourceBase {
// variables have new values that need to be written back.
xla::ComputationDataHandle initial_value;
};
+
+ // Creates a variable with variable `variable_id` and initial type `type` and
+ // value `handle`. `name` is a descriptive name for use in error messages.
+ // Fails if the variable already exists.
+ Status CreateVariable(int variable_id, string name, DataType type,
+ const xla::ComputationDataHandle& handle);
+
+ // Retrieves variable `variable_id`. Fails if the variable does not exist.
+ Status GetVariable(int variable_id, Variable** variable);
+
const std::unordered_map<int, Variable>& variables() { return variables_; }
// Get an XLA lambda to compute Max. This is cached in the
diff --git a/tensorflow/compiler/tf2xla/xla_op_kernel.cc b/tensorflow/compiler/tf2xla/xla_op_kernel.cc
index 4c8c2527bd..f51adba617 100644
--- a/tensorflow/compiler/tf2xla/xla_op_kernel.cc
+++ b/tensorflow/compiler/tf2xla/xla_op_kernel.cc
@@ -206,7 +206,51 @@ Status XlaOpKernelContext::ReadVariableInput(
const Tensor& tensor = context_->input(index);
const XlaExpression* expression = CastExpressionFromTensor(tensor);
int variable_id = expression->variable_id();
- return XlaContext::Get(this).ReadVariable(variable_id, value);
+
+ XlaContext::Variable* variable;
+ XlaContext& context = XlaContext::Get(this);
+ TF_RETURN_IF_ERROR(context.GetVariable(variable_id, &variable));
+ if (variable->value.handle() == 0) {
+ return errors::InvalidArgument("Read of uninitialized variable ",
+ variable->name);
+ }
+ *value = variable->value;
+ return Status::OK();
+}
+
+string XlaOpKernelContext::VariableDebugString(int index) {
+ const Tensor& tensor = context_->input(index);
+ const XlaExpression* expression = CastExpressionFromTensor(tensor);
+ int variable_id = expression->variable_id();
+
+ XlaContext::Variable* variable;
+ XlaContext& context = XlaContext::Get(this);
+ if (!context.GetVariable(variable_id, &variable).ok()) {
+ return "<invalid variable ID>";
+ }
+ return variable->name;
+}
+
+Status XlaOpKernelContext::GetVariableTypeAndShape(int index, DataType* type,
+ TensorShape* shape) const {
+ const Tensor& tensor = context_->input(index);
+ const XlaExpression* expression = CastExpressionFromTensor(tensor);
+ int variable_id = expression->variable_id();
+
+ XlaContext::Variable* variable;
+ XlaContext& context = XlaContext::Get(this);
+ TF_RETURN_IF_ERROR(context.GetVariable(variable_id, &variable));
+ if (variable->value.handle() == 0) {
+ return errors::InvalidArgument("Read of uninitialized variable ",
+ variable->name);
+ }
+ *type = variable->type;
+ auto shape_or_status = builder()->GetShape(variable->value);
+ if (!shape_or_status.ok()) {
+ return shape_or_status.status();
+ }
+ *shape = XLAShapeToTensorShape(*shape_or_status.ValueOrDie());
+ return Status::OK();
}
void XlaOpKernelContext::SetOutput(int index,
@@ -272,7 +316,17 @@ Status XlaOpKernelContext::AssignVariable(
const XlaExpression* expression =
CastExpressionFromTensor(context_->input(index));
XlaContext& context = XlaContext::Get(this);
- return context.AssignVariable(expression->variable_id(), type, handle);
+ XlaContext::Variable* variable;
+ TF_RETURN_IF_ERROR(context.GetVariable(expression->variable_id(), &variable));
+ if (!((variable->type == DT_INVALID && type != DT_INVALID) ||
+ (variable->type == type))) {
+ return errors::InvalidArgument(
+ "Types of variables cannot change after initialization: old type was ",
+ DataTypeString(variable->type), ", new type is ", DataTypeString(type));
+ }
+ variable->type = type;
+ variable->value = handle;
+ return Status::OK();
}
void XlaOpKernelContext::SetOpHasSideEffects() {
diff --git a/tensorflow/compiler/tf2xla/xla_op_kernel.h b/tensorflow/compiler/tf2xla/xla_op_kernel.h
index 8ab9498186..badc8e2274 100644
--- a/tensorflow/compiler/tf2xla/xla_op_kernel.h
+++ b/tensorflow/compiler/tf2xla/xla_op_kernel.h
@@ -141,6 +141,11 @@ class XlaOpKernelContext {
// Variables
+ // Sets `*type` and `*shape` to the current type and shape of a variable's
+ // value.
+ Status GetVariableTypeAndShape(int index, DataType* type,
+ TensorShape* shape) const;
+
// Reads the current value of the resouce variable referred to by input
// 'index'.
Status ReadVariableInput(int index, xla::ComputationDataHandle* value);
@@ -154,6 +159,9 @@ class XlaOpKernelContext {
Status AssignVariable(int variable_index, DataType type,
const xla::ComputationDataHandle& handle);
+ // Returns a human-readable debug string describing 'variable_index'.
+ string VariableDebugString(int variable_index);
+
// Helper routines for the OP_REQUIRES macros
void CtxFailure(Status s);
void CtxFailureWithWarning(Status s);
diff --git a/tensorflow/contrib/framework/python/ops/variables.py b/tensorflow/contrib/framework/python/ops/variables.py
index 1c35b2419c..8296f631a4 100644
--- a/tensorflow/contrib/framework/python/ops/variables.py
+++ b/tensorflow/contrib/framework/python/ops/variables.py
@@ -165,7 +165,7 @@ def local_variable(initial_value, validate_shape=True, name=None):
def variable(name, shape=None, dtype=None, initializer=None,
regularizer=None, trainable=True, collections=None,
caching_device=None, device=None,
- partitioner=None, custom_getter=None):
+ partitioner=None, custom_getter=None, use_resource=None):
"""Gets an existing variable with these parameters or creates a new one.
Args:
@@ -190,6 +190,7 @@ def variable(name, shape=None, dtype=None, initializer=None,
partitions for each axis (currently only one axis can be partitioned).
custom_getter: Callable that allows overwriting the internal
get_variable method and has to have the same signature.
+ use_resource: If `True` use a ResourceVariable instead of a Variable.
Returns:
The created or existing variable.
@@ -209,14 +210,15 @@ def variable(name, shape=None, dtype=None, initializer=None,
trainable=trainable,
collections=collections,
caching_device=caching_device,
- partitioner=partitioner)
+ partitioner=partitioner,
+ use_resource=use_resource)
@contrib_add_arg_scope
def model_variable(name, shape=None, dtype=dtypes.float32, initializer=None,
regularizer=None, trainable=True, collections=None,
caching_device=None, device=None, partitioner=None,
- custom_getter=None):
+ custom_getter=None, use_resource=None):
"""Gets an existing model variable with these parameters or creates a new one.
Args:
@@ -242,6 +244,7 @@ def model_variable(name, shape=None, dtype=dtypes.float32, initializer=None,
partitions for each axis (currently only one axis can be partitioned).
custom_getter: Callable that allows overwriting the internal
get_variable method and has to have the same signature.
+ use_resource: If `True` use a ResourceVariable instead of a Variable.
Returns:
The created or existing variable.
@@ -252,7 +255,8 @@ def model_variable(name, shape=None, dtype=dtypes.float32, initializer=None,
initializer=initializer, regularizer=regularizer,
trainable=trainable, collections=collections,
caching_device=caching_device, device=device,
- partitioner=partitioner, custom_getter=custom_getter)
+ partitioner=partitioner, custom_getter=custom_getter,
+ use_resource=use_resource)
return var
diff --git a/tensorflow/contrib/layers/python/layers/layers.py b/tensorflow/contrib/layers/python/layers/layers.py
index 03cb86601f..8fa734b089 100644
--- a/tensorflow/contrib/layers/python/layers/layers.py
+++ b/tensorflow/contrib/layers/python/layers/layers.py
@@ -1301,7 +1301,8 @@ def _inner_flatten(inputs, new_rank, output_collections=None, scope=None):
def _model_variable_getter(getter, name, shape=None, dtype=None,
initializer=None, regularizer=None, trainable=True,
collections=None, caching_device=None,
- partitioner=None, rename=None, **_):
+ partitioner=None, rename=None, use_resource=None,
+ **_):
"""Getter that uses model_variable for compatibility with core layers."""
short_name = name.split('/')[-1]
if rename and short_name in rename:
@@ -1312,7 +1313,7 @@ def _model_variable_getter(getter, name, shape=None, dtype=None,
name, shape=shape, dtype=dtype, initializer=initializer,
regularizer=regularizer, collections=collections, trainable=trainable,
caching_device=caching_device, partitioner=partitioner,
- custom_getter=getter)
+ custom_getter=getter, use_resource=use_resource)
def _build_variable_getter(rename=None):
diff --git a/tensorflow/python/framework/function_test.py b/tensorflow/python/framework/function_test.py
index 402616db8f..bfe87a9869 100644
--- a/tensorflow/python/framework/function_test.py
+++ b/tensorflow/python/framework/function_test.py
@@ -995,7 +995,9 @@ class VariableHoistingTest(test.TestCase):
self._testSimpleModel(True)
self._testSimpleModel(False)
- def testBasicResource(self):
+ # TODO(b/35668241): disabled because resource variable handling inside
+ # functions does not work.
+ def DISABLED_testBasicResource(self):
self._testSimpleModel(True, use_resource=True)
self._testSimpleModel(False, use_resource=True)
diff --git a/tensorflow/python/ops/variable_scope.py b/tensorflow/python/ops/variable_scope.py
index 47aeca32c3..2a89921944 100644
--- a/tensorflow/python/ops/variable_scope.py
+++ b/tensorflow/python/ops/variable_scope.py
@@ -346,7 +346,7 @@ class _VariableStore(object):
initializer=initializer, regularizer=regularizer,
reuse=reuse, trainable=trainable, collections=collections,
caching_device=caching_device, partitioner=partitioner,
- validate_shape=validate_shape)
+ validate_shape=validate_shape, use_resource=use_resource)
else:
return _true_getter(
name, shape=shape, dtype=dtype,
diff --git a/tensorflow/compiler/jit/kernels/xla_device_launch_op.cc b/tensorflow/compiler/jit/kernels/xla_device_launch_op.cc
index a70a7921e6..c741ccfb31 100644
--- a/tensorflow/compiler/jit/kernels/xla_device_launch_op.cc
+++ b/tensorflow/compiler/jit/kernels/xla_device_launch_op.cc
@@ -16,7 +16,6 @@ limitations under the License.
#include "tensorflow/compiler/jit/kernels/xla_device_launch_op.h"
#include "tensorflow/compiler/jit/defs.h"
-#include "tensorflow/compiler/jit/xla_compilation_cache.h"
#include "tensorflow/compiler/jit/xla_device.h"
#include "tensorflow/compiler/jit/xla_device_context.h"
#include "tensorflow/compiler/xla/client/local_client.h"
@@ -67,20 +66,16 @@ XlaDeviceLaunchOp::XlaDeviceLaunchOp(OpKernelConstruction* ctx)
OP_REQUIRES_OK(ctx, ctx->GetAttr("Nresources", &num_resource_args_));
}
-// Takes a snapshot of the values of resource variable arguments, which are
-// the last `num_variables` arguments. We snapshot tensors that back
-// resource variables since concurrent updates may modify the shape, and it is
-// important that the shapes used for compilation match the true shapes of the
-// buffers.
-static std::vector<OptionalTensor> SnapshotResourceVariables(
- OpKernelContext* ctx, int num_variables) {
+std::vector<OptionalTensor> SnapshotResourceVariables(OpKernelContext* ctx,
+ int num_variables) {
std::vector<OptionalTensor> snapshot(num_variables);
int first_variable = ctx->num_inputs() - num_variables;
for (int i = 0; i < num_variables; ++i) {
Var* variable = nullptr;
- if (LookupResource(ctx, HandleFromInput(ctx, first_variable + i), &variable)
- .ok()) {
+ ResourceHandle handle = HandleFromInput(ctx, first_variable + i);
+ if (LookupResource(ctx, handle, &variable).ok()) {
mutex_lock lock(*variable->mu());
+ snapshot[i].name = handle.name();
snapshot[i].present = true;
snapshot[i].value = *variable->tensor();
}
@@ -127,13 +122,13 @@ void XlaDeviceLaunchOp::Compute(OpKernelContext* ctx) {
// Builds the inputs to the computation.
std::vector<std::shared_ptr<xla::GlobalData>> arg_handles(
- kernel->xla_input_shapes.size());
- std::vector<xla::GlobalData*> arg_ptrs(kernel->xla_input_shapes.size());
+ kernel->input_mapping.size());
+ std::vector<xla::GlobalData*> arg_ptrs(kernel->input_mapping.size());
// Adds the argument tensors.
const int first_variable_arg = ctx->num_inputs() - num_resource_args_;
- for (int i = 0; i < kernel->xla_input_shapes.size(); ++i) {
- int op_input_num = kernel->xla_input_shapes[i].first;
+ for (int i = 0; i < kernel->input_mapping.size(); ++i) {
+ int op_input_num = kernel->input_mapping[i];
if (op_input_num >= first_variable_arg) {
arg_handles[i] = XlaTransferManager::GetTensorGlobalData(
@@ -201,10 +196,10 @@ void XlaDeviceLaunchOp::Compute(OpKernelContext* ctx) {
}
}
- // Apply variable writes, if any.
- VLOG(2) << "Applying variable writes";
- for (int i = 0; i < kernel->variable_writes.size(); ++i) {
- const XlaCompiler::VariableWrite& write = kernel->variable_writes[i];
+ // Apply variable updates, if any.
+ VLOG(2) << "Applying variable updates";
+ for (int i = 0; i < kernel->variable_updates.size(); ++i) {
+ const XlaCompiler::VariableUpdate& write = kernel->variable_updates[i];
OP_REQUIRES(ctx,
write.input_index >= 0 && write.input_index < ctx->num_inputs(),
errors::Internal("Invalid input index for variable write."));
diff --git a/tensorflow/compiler/tf2xla/xla_compiler.h b/tensorflow/compiler/tf2xla/xla_compiler.h
index 477802c6a7..3ed920521b 100644
--- a/tensorflow/compiler/tf2xla/xla_compiler.h
+++ b/tensorflow/compiler/tf2xla/xla_compiler.h
@@ -34,15 +34,48 @@ namespace tensorflow {
// It does a symbolic execution of the graph starting from specific input
// shapes, using a JIT device to convert operators into XLA computations.
//
-// It is typically invoked from an `_XlaLaunch` operator once the shapes
-// of all input parameters to the computation are known. This is
+// XlaCompiler is typically invoked from an `_XlaLaunch` operator once the
+// shapes of all input parameters to the computation are known. This is
// because the symbolic execution requires known shapes for all operations.
+//
+// XlaCompiler compiles Tensorflow graphs that received inputs via _Arg nodes,
+// and return outputs via _Retval nodes.
+//
+// The XlaCompiler requires one Argument struct for each _Arg index, that
+// describes each argument. Arguments can be compile-time constants
+// (kind kConstant), run-time parameters (kind kParameter), or resource
+// variables (kinds kVariable and kUninitializedVariable).
+//
+// Only kParameter and kVariable arguments become runtime parameters to the
+// generated XLA computation. The XLA computation will have run-time parameters
+// in the following order:
+// +---------------------+-----------------------------------------+
+// | kParameter values | Initial values of kVariable arguments |
+// +---------------------+-----------------------------------------+
+// Within each block, the arguments are arranged by the _Arg index from which
+// they were derived.
+// If `Options::requires_runtime_context` is true, then an additional runtime
+// context argument is passed as a final argument.
+//
+// The run-time outputs of the XLA computation are arranged in the following
+// order:
+// +------------------+-----------------------------------------+
+// | _Retval values | Updated values of kVariable arguments |
+// +------------------+-----------------------------------------+
+// _Retval values are ordered by _Retval index, whereas kVariable values are
+// ordered by the original _Arg position of the variable.
+//
+// In both inputs and outputs, kVariable values are placed the end. When
+// emitting While loop bodies, we must ensure that the loop body has
+// identical input and output signatures. By moving variable values
+// to the end of the argument list and using the
+// `return_updated_values_for_all_variables` option, we can ensure that the
+// input and output values of variables appear at the same positions.
+
class XlaCompiler {
public:
// Describes how to derive the value of each _Arg node in the graph/function
- // being compiled. Each argument must be either a parameter of the generated
- // XLA computation (parameter >= 0), or a compile time constant
- // (parameter < 0).
+ // being compiled. There must be one Argument for each _Arg index.
struct Argument {
enum Kind {
// Default value; not a valid kind.
@@ -82,7 +115,8 @@ class XlaCompiler {
};
struct OutputDescription {
- // Shape of the output.
+ // Type and shape of the output.
+ DataType type;
TensorShape shape;
// Constant output value, if known to be constant at JIT compilation time.
@@ -92,28 +126,38 @@ class XlaCompiler {
};
// Describes a variable write side effect of the computation.
- struct VariableWrite {
+ struct VariableUpdate {
// Index of the input that contains the variable resource to write to.
int input_index;
// Type and shape of the tensor to be written back.
DataType type;
TensorShape shape;
+
+ // Was the value of the variable modified by the computation?
+ // (Always true, unless `return_updated_values_for_all_variables` is true.)
+ bool modified;
};
struct CompilationResult {
- // Vector of (Tensorflow input number, XLA shape) pairs that describe
- // the arguments of the compiled XLA computation. (Because of constant
- // inputs, the arguments to the XLA computation are a subset of the
- // inputs passed to the JIT.)
- std::vector<std::pair<int, xla::Shape>> xla_input_shapes;
+ // Vector that maps from the parameters of the XLA computation to their
+ // original argument positions. To handle compile-time constant inputs and
+ // variables, the parameters to the XLA computation may be a subset of the
+ // original arguments, and are not necessarily in the same order.)
+ std::vector<int> input_mapping;
// Does the computation require the local runtime context to be passed as
// the last argument?
bool requires_runtime_context = false;
- // Output shape in XLA format. This is a tuple if and only if
- // there are multiple non-constant outputs.
+ // Input shapes of the computation.
+ std::vector<xla::Shape> xla_input_shapes;
+
+ // Should the arguments be packed into a single tuple?
+ bool tuple_arg;
+
+ // Output shape in XLA format. The output shape is a tuple if and only if
+ // the number of non-constant outputs is not equal to 1.
xla::Shape xla_output_shape;
// TensorFlow shapes of outputs, together with the values of any
@@ -121,10 +165,10 @@ class XlaCompiler {
// containing both constant and non-constant results.
std::vector<OutputDescription> outputs;
- // Variables whose values should be written by the computation back, ordered
- // by return value position. Variable write results follow the non-constant
+ // Variables whose values were updated by the computation, ordered
+ // by return value position. Variable updates follow the non-constant
// results in the outputs of XLA computation.
- std::vector<VariableWrite> variable_writes;
+ std::vector<VariableUpdate> variable_updates;
// The XLA computation built from the tensorflow subgraph. May be null
// if the output consists solely of compile-time constants.
@@ -153,21 +197,38 @@ class XlaCompiler {
// as Tensors at compile-time, rather than as run-time outputs of the
// computation.
bool resolve_compile_time_constants = true;
+
+ // If `use_tuple_arg` is true, a single tuple parameter will be used for all
+ // arguments; if false, each argument gets its own parameter.
+ bool use_tuple_arg = false;
+
+ // If 'return_updated_values_for_all_variables' is true, then updated
+ // values of all resource variables arguments will be included in the
+ // 'variable_updates' of the computation, even if the variable was not
+ // modified by the computation. Used when compiling loop bodies to ensure
+ // the input and output signatures match.
+ bool return_updated_values_for_all_variables = false;
+
+ // If 'prune_unreachable_nodes' is true, then nodes that are not
+ // dependencies of graph's _Retval nodes will be pruned before compilation.
+ // This is useful to prune stateful operators that should not be executed
+ // from a function body.
+ bool prune_unreachable_nodes = false;
};
- explicit XlaCompiler(const Options& options);
+ explicit XlaCompiler(Options options);
~XlaCompiler();
// Compiles a Tensorflow function `fn_name_attrs` into an XLA computation.
// `args` describes the arguments to the function, each of which must either
- // be a parameter to the XLA computation or a compile-time constant.
- // Writes the compiled output to `result`.
+ // be a runtime-parameter to the XLA computation, a compile-time constant, or
+ // a resource variable. Writes the compiled output to `result`.
//
// The generated XLA computation returns a tuple containing only the
// non-constant outputs as a function of the input arguments. Constant
// arguments are returned as host memory tensors in the output list and are
// not included in the XLA computation's outputs. The XLA computation is
- // null if there are no data-dependent outputs.
+ // null if there are no data-dependent outputs and no side effects.
Status CompileFunction(FunctionLibraryRuntime* flr,
const NameAttrList& fn_name_attrs,
const std::vector<Argument>& args,
@@ -176,41 +237,17 @@ class XlaCompiler {
// Compiles a tensorflow::Graph into an xla::Computation.
// Similar to CompileFunction, but takes a Graph as input rather than a
// function.
- // If `use_tuple_arg` is true, the compilation takes all of its arguments as
- // a single tuple.
Status CompileGraph(string const& name, std::unique_ptr<Graph> graph,
FunctionLibraryRuntime* flr,
- const std::vector<Argument>& args, bool use_tuple_arg,
+ const std::vector<Argument>& args,
CompilationResult* result);
- // Helper function that compiles a function to an XLA computation suitable
- // for use as a subroutine in other Computations, e.g., the body of a
- // While loop.
- //
- // The emitted Computation takes a single input parameter with
- // input_shape. If this is a tuple then the tuple element shapes
- // must match the types of the function's _Arg nodes. If input_shape
- // is not a tuple then the function must have a single _Arg node
- // with the same type as input_shape. The shapes of the _Arg values
- // will be compiled to match input_shape.
- //
- // The emitted Computation also returns a single value. If output_shape is a
- // tuple the tuple elements' types and shapes must match the compiled
- // function's _Retval nodes. If output_shape is not a tuple the
- // function must have a single _Retval node with the correct type
- // (and shape after compilation).
- Status CompileSubComputation(FunctionLibraryRuntime* flr,
- const NameAttrList& fn_name_attrs,
- const xla::Shape& input_shape,
- const xla::Shape& output_shape,
- xla::Computation* computation);
-
- // Takes <*result>, which has been compiled from a Tensorflow subgraph to a
+ // Takes `result` which has been compiled from a Tensorflow subgraph to a
// XLA computation already, and generates an XLA LocalExecutable `executable`.
Status BuildExecutable(const CompilationResult& result,
std::unique_ptr<xla::LocalExecutable>* executable);
- xla::Client* client() const { return client_; }
+ xla::Client* client() const { return options_.client; }
XlaCompilationDevice* device() const { return device_; }
const DeviceMgr* device_mgr() const { return &device_mgr_; }
@@ -221,17 +258,7 @@ class XlaCompiler {
Status GetChannelHandle(const string& key, xla::ChannelHandle* channel);
private:
- // Does the real work of Compile() and CompileToComputation().
- Status CompileFunctionBody(FunctionLibraryRuntime* flr,
- const FunctionBody& function_body,
- const string& name,
- const std::vector<Argument>& args,
- bool use_tuple_arg, CompilationResult* result);
-
- xla::Client* client_; // Not owned.
- const bool allow_cpu_custom_calls_;
- const bool local_executable_has_hybrid_result_;
- const bool resolve_compile_time_constants_;
+ Options options_;
// Returns the next step sequence number.
int64 NextStepId();